import logging
import mimetypes

from django.contrib.staticfiles.finders import find
from django.contrib.staticfiles.storage import staticfiles_storage
from django.template.loader import get_template
from weasyprint import default_url_fetcher, HTML

from . import models

logger = logging.getLogger(__name__)


def staticfile_url_fetcher(url):
    """
    Returns files when the staticfiles app does not returns a relative path to
    a file.
    """
    if url.startswith('/'):
        base_url = staticfiles_storage.base_url
        filename = url.replace(base_url, '', 1)

        filepath = find(filename)
        with open(filepath, 'rb') as file_:
            data = file_.read()
        return {
            'string': data,
            'mime_type': mimetypes.guess_type(url)[0],
        }
    else:
        return default_url_fetcher(url)


def generate_receipt_pdf(pk, target, force_html=False):

    pdf = models.ReceiptPDF.objects.select_related(
        'receipt',
        'receipt__receipt_type',
        'receipt__document_type',
        'receipt__validation',
        'receipt__point_of_sales',
        'receipt__point_of_sales__owner',
    ).prefetch_related(
        'receipt__entries',
    ).get(
        receipt__pk=pk
    )

    html = get_template('receipts/code_{}.html'.format(
        pdf.receipt.receipt_type.code,
    )).render({
        'pdf': pdf,
        'taxpayer': pdf.receipt.point_of_sales.owner,
    })

    if force_html:
        return html
    else:
        return HTML(
            string=html,
            base_url='not-used://',
            url_fetcher=staticfile_url_fetcher,
        ).write_pdf(
            target=target,
        )



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0004_auto_20150916_1934'),
    ]

    operations = [
        migrations.AlterField(
            model_name='receipt',
            name='point_of_sales',
            field=models.ForeignKey(to='afip.PointOfSales', verbose_name='punto de ventas', related_name='receipts', on_delete=models.CASCADE),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='receipt_type',
            field=models.ForeignKey(to='afip.ReceiptType', verbose_name='tipo de comprobante', related_name='receipts', on_delete=models.CASCADE),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='related_receipts',
            field=models.ManyToManyField(blank=True, to='afip.Receipt', verbose_name='comprobantes relacionados'),
        ),
        migrations.AlterUniqueTogether(
            name='pointofsales',
            unique_together=set([('number', 'owner')]),
        ),
    ]



import io
import logging
import random
import uuid
from base64 import b64encode
from datetime import datetime, timedelta, timezone
from tempfile import NamedTemporaryFile

import pytz
from django.core.files.base import File
from django.db import models
from django.utils.translation import ugettext_lazy as _
from lxml import etree
from lxml.builder import E
from zeep.exceptions import Fault

from . import clients, crypto, exceptions, parsers, serializers

logger = logging.getLogger(__name__)
TZ_AR = pytz.timezone(pytz.country_timezones['ar'][0])


def populate_all():
    """Fetch and store all metadata from the AFIP."""
    ReceiptType.objects.populate()
    ConceptType.objects.populate()
    DocumentType.objects.populate()
    VatType.objects.populate()
    TaxType.objects.populate()
    CurrencyType.objects.populate()


def check_response(response):
    """
    Check that a response is not an error.

    AFIP allows us to create valid tickets with invalid key/CUIT pairs, so we
    can end up with tickets that fail on any service.

    Due to how zeep works, we can't quite catch these sort of errors on some
    middleware level (well, honestly, we need to do a large refactor).

    This method checks if responses have an error, and raise a readable
    message.
    """
    if response.Errors:
        raise exceptions.AfipException(response)


class GenericAfipTypeManager(models.Manager):
    """Default Manager for GenericAfipType."""

    def __init__(self, service_name, type_name):
        """
        Create a new Manager instance for a GenericAfipType.

        This should generally only be required to manually populate a single
        type with upstream data.
        """
        super().__init__()
        self.__service_name = service_name
        self.__type_name = type_name

    def populate(self, ticket=None):
        """
        Populate the database with types retrieved from the AFIP.

        If no ticket is provided, the most recent available one will be used.
        """
        ticket = ticket or AuthTicket.objects.get_any_active('wsfe')
        client = clients.get_client('wsfe', ticket.owner.is_sandboxed)
        service = getattr(client.service, self.__service_name)
        response_xml = service(serializers.serialize_ticket(ticket))

        check_response(response_xml)

        for result in getattr(response_xml.ResultGet, self.__type_name):
            self.get_or_create(
                code=result.Id,
                description=result.Desc.encode('UTF-8'),
                valid_from=parsers.parse_date(result.FchDesde),
                valid_to=parsers.parse_date(result.FchHasta),
            )


class GenericAfipType(models.Model):
    """An abstract class for several of AFIP's metadata types."""

    code = models.CharField(
        _('code'),
        max_length=3,
    )
    description = models.CharField(
        _('description'),
        max_length=250,
    )
    valid_from = models.DateField(
        _('valid from'),
        null=True,
        blank=True,
    )
    valid_to = models.DateField(
        _('valid until'),
        null=True,
        blank=True,
    )

    def __str__(self):
        return self.description

    class Meta:
        abstract = True


class ReceiptType(GenericAfipType):
    """
    An AFIP receipt type.

    See the AFIP's documentation for details on each receipt type.
    """

    objects = GenericAfipTypeManager('FEParamGetTiposCbte', 'CbteTipo')

    class Meta:
        verbose_name = _('receipt type')
        verbose_name_plural = _('receipt types')


class ConceptType(GenericAfipType):
    """
    An AFIP concept type.

    See the AFIP's documentation for details on each concept type.
    """

    objects = GenericAfipTypeManager('FEParamGetTiposConcepto', 'ConceptoTipo')

    class Meta:
        verbose_name = _('concept type')
        verbose_name_plural = _('concept types')


class DocumentType(GenericAfipType):
    """
    An AFIP document type.

    See the AFIP's documentation for details on each document type.
    """

    objects = GenericAfipTypeManager('FEParamGetTiposDoc', 'DocTipo')

    class Meta:
        verbose_name = _('document type')
        verbose_name_plural = _('document types')


class VatType(GenericAfipType):
    """
    An AFIP VAT type.

    See the AFIP's documentation for details on each VAT type.
    """

    objects = GenericAfipTypeManager('FEParamGetTiposIva', 'IvaTipo')

    class Meta:
        verbose_name = _('vat type')
        verbose_name_plural = _('vat types')


class TaxType(GenericAfipType):
    """
    An AFIP tax type.

    See the AFIP's documentation for details on each tax type.
    """

    objects = GenericAfipTypeManager('FEParamGetTiposTributos', 'TributoTipo')

    class Meta:
        verbose_name = _('tax type')
        verbose_name_plural = _('tax types')


class CurrencyType(GenericAfipType):
    """
    An AFIP curreny type.

    See the AFIP's documentation for details on each currency type.
    """

    objects = GenericAfipTypeManager('FEParamGetTiposMonedas', 'Moneda')

    def __str__(self):
        return self.code

    class Meta:
        verbose_name = _('currency type')
        verbose_name_plural = _('currency types')


class TaxPayer(models.Model):
    """
    Represents an AFIP TaxPayer.

    This class has the bare minimum required for most AFIP services.

    Note that multiple instances of this object can actually represent the same
    taxpayer, each using a different key.
    """

    name = models.CharField(
        _('name'),
        max_length=32,
        help_text=_('A friendly name to recognize this taxpayer.'),
    )
    key = models.FileField(
        _('key'),
        null=True,
    )
    certificate = models.FileField(
        _('certificate'),
        null=True,
    )
    cuit = models.BigIntegerField(
        _('cuit'),
    )
    is_sandboxed = models.BooleanField(
        _('is sandboxed'),
        help_text=_(
            'Indicates if this taxpayer interacts with the sandbox servers '
            'rather than the production servers'
        )
    )

    def generate_key(self, force=False):
        """
        Creates a key file for this TaxPayer

        Creates a key file for this TaxPayer if it does not have one, and
        immediately saves it.

        Returns True if and only if a key was created.
        """
        if self.key and not force:
            logger.warning(
                'Tried to generate key for a taxpayer that already had one'
            )
            return False

        with NamedTemporaryFile(suffix='.key') as file_:
            crypto.create_key(file_)
            self.key = File(file_, name='{}.key'.format(uuid.uuid4().hex))
            self.save()

        return True

    def generate_csr(self, basename='djangoafip'):
        """
        Creates a CSR for this TaxPayer's key

        Creates a file-like object that contains the CSR which can be used to
        request a new certificate from AFIP.
        """
        csr = io.BytesIO()
        crypto.create_csr(
            self.key.file,
            self.name,
            '{}{}'.format(basename, int(datetime.now().timestamp())),
            'CUIT {}'.format(self.cuit),
            csr,
        )
        csr.seek(0)
        return csr

    def create_ticket(self, service):
        """Create an AuthTicket for a given service."""
        ticket = AuthTicket(owner=self, service=service)
        ticket.authorize()
        return ticket

    def get_ticket(self, service):
        """Return an existing AuthTicket for a given service."""
        return self.auth_tickets \
            .filter(expires__gt=datetime.now(timezone.utc), service=service) \
            .last()

    def get_or_create_ticket(self, service):
        """
        Return or create a new AuthTicket for a given serivce.

        Return an existing ticket for a service if one is available, otherwise,
        create a new one and return that.

        This is generally the preferred method of obtaining tickets for any
        service.
        """
        return self.get_ticket(service) or self.create_ticket(service)

    def fetch_points_of_sales(self, ticket=None):
        """
        Fetch all point of sales objects.

        Fetch all point of sales from the WS and store (or update) them
        locally.

        Returns a list of tuples with the format (pos, created,).
        """
        ticket = ticket or self.get_or_create_ticket('wsfe')

        client = clients.get_client('wsfe', self.is_sandboxed)
        response = client.service.FEParamGetPtosVenta(
            serializers.serialize_ticket(ticket),
        )
        check_response(response)

        results = []
        for pos_data in response.ResultGet.PtoVenta:
            results.append(PointOfSales.objects.update_or_create(
                number=pos_data.Nro,
                issuance_type=pos_data.EmisionTipo,
                owner=self,
                defaults={
                    'blocked': pos_data.Bloqueado == 'N',
                    'drop_date': parsers.parse_date(pos_data.FchBaja),
                }
            ))

        return results

    def __repr__(self):
        return '<TaxPayer {}: {}, CUIT {}>'.format(
            self.pk,
            self.name,
            self.cuit,
        )

    def __str__(self):
        return str(self.cuit)

    class Meta:
        verbose_name = _('taxpayer')
        verbose_name_plural = _('taxpayers')


class TaxPayerProfile(models.Model):
    """
    Custom information about a taxpayer, used in printed receipts.

    Most of these can be overriden per-receipt, and are usually just defaults.

    None of these are required or sent to the AFIP when notifying about receipt
    generation. They are used *only* for PDF generation.
    """

    taxpayer = models.OneToOneField(
        TaxPayer,
        related_name='profile',
        verbose_name=_('taxpayer'),
        on_delete=models.CASCADE,
    )
    issuing_name = models.CharField(
        max_length=128,
        verbose_name=_('issuing name'),
    )
    issuing_address = models.TextField(
        _('issuing address'),
    )
    issuing_email = models.CharField(
        max_length=128,
        verbose_name=_('issuing email'),
        blank=True,
        null=True,
    )
    vat_condition = models.CharField(
        max_length=48,
        verbose_name=_('vat condition'),
    )
    gross_income_condition = models.CharField(
        max_length=48,
        verbose_name=_('gross income condition'),
    )
    sales_terms = models.CharField(
        max_length=48,
        verbose_name=_('sales terms'),
        help_text=_(
            'The terms of the sale printed onto receipts by default '
            '(eg: single payment, checking account, etc).'
        ),
    )
    active_since = models.DateField(
        _('active since'),
        help_text=_(
            'Date since which this taxpayer has been legally active.'
        ),
    )

    class Meta:
        verbose_name = _('taxpayer profile')
        verbose_name_plural = _('taxpayer profiles')


class PointOfSales(models.Model):
    """
    Represents an existing AFIP point of sale.

    Points of sales need to be created via AFIP's web interface and it is
    recommended that you use :meth:`~.TaxPayer.fetch_points_of_sales` to fetch
    these programatically.
    """
    number = models.PositiveSmallIntegerField(
        _('number'),
    )
    issuance_type = models.CharField(
        _('issuance type'),
        max_length=8,
        help_text='Indicates if thie POS emits using CAE and CAEA.'
    )
    blocked = models.BooleanField(
        _('blocked'),
    )
    drop_date = models.DateField(
        _('drop date'),
        null=True,
        blank=True,
    )

    owner = models.ForeignKey(
        TaxPayer,
        related_name='points_of_sales',
        verbose_name=_('owner'),
        on_delete=models.CASCADE,
    )

    def __str__(self):
        return str(self.number)

    class Meta:
        unique_together = (
            ('number', 'owner'),
        )
        verbose_name = _('point of sales')
        verbose_name_plural = _('points of sales')


class AuthTicketManager(models.Manager):

    def get_any_active(self, service):
        ticket = AuthTicket.objects.filter(
            token__isnull=False,
            expires__gt=datetime.now(timezone.utc),
            service=service,
        ).first()
        if ticket:
            return ticket

        taxpayer = TaxPayer.objects.order_by('?').first()

        if not taxpayer:
            raise exceptions.AuthenticationError(
                _('There are no taxpayers to generate a ticket.'),
            )

        return taxpayer.create_ticket(service)


class AuthTicket(models.Model):
    """
    An AFIP Authorization ticket.

    This is a signed ticket used to communicate with AFIP's webservices.

    Applications should not generally have to deal with these tickets
    themselves; most services will find or create one as necessary.
    """

    def default_generated():
        return datetime.now(TZ_AR)

    def default_expires():
        tomorrow = datetime.now(TZ_AR) + timedelta(hours=12)
        return tomorrow

    def default_unique_id():
        return random.randint(0, 2147483647)

    owner = models.ForeignKey(
        TaxPayer,
        verbose_name=_('owner'),
        related_name='auth_tickets',
        on_delete=models.CASCADE,
    )
    unique_id = models.IntegerField(
        _('unique id'),
        default=default_unique_id,
    )
    generated = models.DateTimeField(
        _('generated'),
        default=default_generated,
    )
    expires = models.DateTimeField(
        _('expires'),
        default=default_expires,
    )
    service = models.CharField(
        _('service'),
        max_length=6,
        help_text=_('Service for which this ticket has been authorized'),
    )

    token = models.TextField(
        _('token'),
    )
    signature = models.TextField(
        _('signature'),
    )

    objects = AuthTicketManager()

    TOKEN_XPATH = '/loginTicketResponse/credentials/token'
    SIGN_XPATH = '/loginTicketResponse/credentials/sign'

    def __create_request_xml(self):
        request_xml = (
            E.loginTicketRequest(
                {'version': '1.0'},
                E.header(
                    E.uniqueId(str(self.unique_id)),
                    E.generationTime(
                        serializers.serialize_datetime(self.generated)
                    ),
                    E.expirationTime(
                        serializers.serialize_datetime(self.expires)
                    ),
                ),
                E.service(self.service)
            )
        )
        return etree.tostring(request_xml, pretty_print=True)

    def __sign_request(self, request):
        with open(self.owner.certificate.file.name) as cert_file:
            cert = cert_file.read()

        with open(self.owner.key.file.name) as key_file:
            key = key_file.read()

        return crypto.create_embeded_pkcs7_signature(request, cert, key)

    def authorize(self):
        """Send this ticket to AFIP for authorization."""
        request = self.__create_request_xml()
        request = self.__sign_request(request)
        request = b64encode(request).decode()

        client = clients.get_client('wsaa', self.owner.is_sandboxed)
        try:
            raw_response = client.service.loginCms(request)
        except Fault as e:
            if e.message == 'Certificado expirado':
                raise exceptions.CertificateExpired(e.message) from e
            if e.message == 'Certificado no emitido por AC de confianza':
                raise exceptions.UntrustedCertificate(e.message) from e
            raise exceptions.AuthenticationError(e.message) from e
        response = etree.fromstring(raw_response.encode('utf-8'))

        self.token = response.xpath(self.TOKEN_XPATH)[0].text
        self.signature = response.xpath(self.SIGN_XPATH)[0].text

        self.save()

    def __str__(self):
        return str(self.unique_id)

    class Meta:
        verbose_name = _('authorization ticket')
        verbose_name_plural = _('authorization tickets')


class ReceiptBatchManager(models.Manager):

    def create(self, queryset):
        """
        Creates a batch with all receipts returned by ``queryset``.

        If you need to create a ReceiptBatch with a single receipt, just
        pass a query that returns that one::

            Receipt.objects.filter(pk=pk)
        """
        first = queryset.select_related('point_of_sales').first()
        if not first:
            # Queryset is empty, nothing to create
            return
        batch = ReceiptBatch(
            receipt_type_id=first.receipt_type_id,
            point_of_sales_id=first.point_of_sales_id,
        )
        batch.save()

        # Exclude any receipts that are already batched (either pre-selection,
        # or due to concurrency):
        queryset.filter(batch__isnull=True).update(batch=batch)
        return batch


class ReceiptBatch(models.Model):
    """
    Receipts are validated in batches.

    AFIP's webservice validates receipts in batches, and this class models
    them. A batch of receipts is simply a set of consecutive-numbered
    :class:`~.Receipt` instances, with the same :class:`~.ReceiptType` and
    :class:`~.PointOfSales`.

    If you need to validate a single Receipt, it's okay to create a batch with
    just one.
    """

    receipt_type = models.ForeignKey(
        ReceiptType,
        verbose_name=_('receipt type'),
        on_delete=models.PROTECT,
    )
    point_of_sales = models.ForeignKey(
        PointOfSales,
        verbose_name=_('point of sales'),
        on_delete=models.PROTECT,
    )

    def __str__(self):
        return str(self.id)

    def validate(self, ticket=None):
        """
        Validates all receipts assigned to this batch.

        Attempting to validate an empty batch will do nothing.

        Any receipts that fail validation are removed from the batch, so you
        should never modify a batch after validation. Receipts that succesfully
        validae will have a :class:`~.ReceiptValidation` object attatched to
        them with a validation date and CAE information.

        Returns a list of errors as returned from AFIP's webservices. An
        exception is not raised because partial failures are possible.
        """
        if self.receipts.count() == 0:
            logger.debug('Refusing to validate empty Batch')
            return []
        if self.receipts.filter(validation__isnull=True).count() == 0:
            logger.debug('Refusing to Batch with no non-validated Receipts')
            return []

        ticket = ticket or \
            self.point_of_sales.owner.get_or_create_ticket('wsfe')

        next_num = Receipt.objects.fetch_last_receipt_number(
            self.point_of_sales,
            self.receipt_type,
        ) + 1
        for receipt in self.receipts.filter(receipt_number__isnull=True):

            # Atomically update receipt number
            Receipt.objects \
                .filter(pk=receipt.id, receipt_number__isnull=True) \
                .update(receipt_number=next_num)
            next_num += 1

        # Purge the internal cache (.update() doesn't maintan it)
        self.receipts.all()

        client = clients.get_client(
            'wsfe', self.point_of_sales.owner.is_sandboxed
        )
        response = client.service.FECAESolicitar(
            serializers.serialize_ticket(ticket),
            serializers.serialize_receipt_batch(self),
        )
        check_response(response)

        if response.Errors:
            raise exceptions.AfipException(response)

        validation = Validation.objects.create(
            processed_date=parsers.parse_datetime(
                response.FeCabResp.FchProceso
            ),
            result=response.FeCabResp.Resultado,
            batch=self,
        )

        errs = []
        for cae_data in response.FeDetResp.FECAEDetResponse:
            if cae_data.Resultado == Validation.RESULT_APPROVED:
                rv = validation.receipts.create(
                    result=cae_data.Resultado,
                    cae=cae_data.CAE,
                    cae_expiration=parsers.parse_date(cae_data.CAEFchVto),
                    receipt=self.receipts.get(
                        receipt_number=cae_data.CbteDesde
                    ),
                )
                if cae_data.Observaciones:
                    for obs in cae_data.Observaciones.Obs:
                        observation = Observation.objects.get_or_create(
                            code=obs.Code,
                            message=obs.Msg,
                        )
                    rv.observations.add(observation)
            elif cae_data.Observaciones:
                for obs in cae_data.Observaciones.Obs:
                    errs.append(
                        'Error {}: {}'.format(
                            obs.Code,
                            parsers.parse_string(obs.Msg),
                        )
                    )

        # Remove failed receipts from the batch
        self.receipts.filter(validation__isnull=True) \
            .update(batch=None, receipt_number=None)

        return errs

    class Meta:
        verbose_name = _('receipt batch')
        verbose_name_plural = _('receipt batches')

    objects = ReceiptBatchManager()


class ReceiptQuerySet(models.QuerySet):
    """
    The default queryset obtains when querying via :class:`~.ReceiptManager`.
    """

    def validate(self, ticket=None):
        """
        Validates all receipts matching this queryset.

        Note that, due to how AFIP implements its numbering, this method is not
        thread-safe, or even multiprocess-safe.

        Because of this, it is possible that not all instances matching this
        queryset are validates properly; however, consistency *is* and receipts
        will be updated if and only if they have been validated by the AFIP.
        """
        return ReceiptBatch.objects.create(self).validate(ticket)


class ReceiptManager(models.Manager):
    """
    The default manager for the :class:`~.Receipt` class.

    You should generally access this using ``Receipt.objects``.
    """

    def fetch_last_receipt_number(self, point_of_sales, receipt_type):
        client = clients.get_client('wsfe', point_of_sales.owner.is_sandboxed)
        response_xml = client.service.FECompUltimoAutorizado(
            serializers.serialize_ticket(
                point_of_sales.owner.get_or_create_ticket('wsfe')
            ),
            point_of_sales.number,
            receipt_type.code,
        )
        check_response(response_xml)

        # TODO XXX: Error handling
        # (FERecuperaLastCbteResponse){
        #    PtoVta = 0
        #    CbteTipo = 0
        #    CbteNro = 0
        #    Errors =
        #       (ArrayOfErr){
        #          Err[] =
        #             (Err){
        #                Code = 601
        #                Msg = "CUIT representada no incluida en Token"
        #             },
        #       }
        #  }

        return response_xml.CbteNro

    def get_queryset(self):
        return ReceiptQuerySet(self.model, using=self._db).select_related(
            'receipt_type',
        )


class Receipt(models.Model):
    """
    A receipt, as sent to AFIP.

    Note that AFIP allows sending ranges of receipts, but this isn't generally
    what you want, so we model invoices individually.

    To validate a Receipt, you need to create a :class:`~.ReceiptBatch` first.

    You'll probably want to relate some `Sale` or `Order` object from your
    model with each Receipt.

    All ``document_`` fields contain the recipient's data.

    If the taxpayer has taxes or pays VAT, you need to attach :class:`~.Tax`
    and/or :class:`~.Vat` instances to the Receipt.
    """
    batch = models.ForeignKey(
        ReceiptBatch,
        related_name='receipts',
        null=True,
        blank=True,
        verbose_name=_('receipt batch'),
        help_text=_(
            'Receipts are validated in batches, so it must be assigned one '
            'before validation is possible.'
        ),
        on_delete=models.PROTECT,
    )
    concept = models.ForeignKey(
        ConceptType,
        verbose_name=_('concept'),
        related_name='receipts',
        on_delete=models.PROTECT,
    )
    document_type = models.ForeignKey(
        DocumentType,
        verbose_name=_('document type'),
        related_name='receipts',
        help_text=_(
            'The document type of the customer to whom this receipt '
            'is addressed'
        ),
        on_delete=models.PROTECT,
    )
    document_number = models.BigIntegerField(
        _('document number'),
        help_text=_(
            'The document number of the customer to whom this receipt '
            'is addressed'
        )
    )
    # NOTE: WS will expect receipt_from and receipt_to.
    receipt_number = models.PositiveIntegerField(
        _('receipt number'),
        null=True,
        blank=True,
        help_text=_(
            'If left blank, the next valid number will assigned when '
            'validating the receipt.'
        )
    )
    issued_date = models.DateField(
        verbose_name=_('issued date'),
        help_text=_('Can diverge up to 5 days for good, or 10 days otherwise'),
    )
    total_amount = models.DecimalField(
        # ImpTotal
        _('total amount'),
        max_digits=15,
        decimal_places=2,
        help_text=_(
            'Must be equal to untaxed amount + exempt amount + taxes + vat.'
        )
    )
    net_untaxed = models.DecimalField(
        # ImpTotConc
        _('total untaxable amount'),
        max_digits=15,
        decimal_places=2,
        help_text=_(
            'The total amount to which taxes do not apply. '
            'For C-type receipts, this must be zero.'
        ),
    )
    net_taxed = models.DecimalField(
        # ImpNeto
        _('total taxable amount'),
        max_digits=15,
        decimal_places=2,
        help_text=_(
            'The total amount to which taxes apply. '
            'For C-type receipts, this is equal to the subtotal.'
            ),
    )
    exempt_amount = models.DecimalField(
        # ImpOpEx
        # Sólo para emisores que son IVA exento
        _('exempt amount'),
        max_digits=15,
        decimal_places=2,
        help_text=_(
            'Only for categories which are tax-exempt. '
            'For C-type receipts, this must be zero.'
        ),
    )
    # Runtime-calculated. Deleteable?
    # vat_amount = models.DecimalField(
    #     _('vat amount'),
    #     max_digits=15,
    #     decimal_places=2,
    #     help_text=_('Must be equal to the sum of all Vat objects.'),
    # )
    # tax_amount = models.DecimalField(
    #     _('tax amount'),
    #     max_digits=15,
    #     decimal_places=2,
    #     help_text=_('Must be equal to the sum of all Tax objects.'),
    # )
    service_start = models.DateField(
        _('service start date'),
        help_text=_(
            'Date on which a service started. No applicable for goods.'
        ),
        null=True,
        blank=True,
    )
    service_end = models.DateField(
        _('service end date'),
        help_text=_(
            'Date on which a service ended. No applicable for goods.'
        ),
        null=True,
        blank=True,
    )
    expiration_date = models.DateField(
        _('receipt expiration date'),
        help_text=_(
            'Date on which this receipt expires. No applicable for goods.'
        ),
        null=True,
        blank=True,
    )
    currency = models.ForeignKey(
        CurrencyType,
        verbose_name=_('currency'),
        related_name='documents',
        help_text=_(
            'Currency in which this receipt is issued.',
        ),
        on_delete=models.PROTECT,
    )
    currency_quote = models.DecimalField(
        _('currency quote'),
        max_digits=10,
        decimal_places=6,
        help_text=_(
            'Quote of the day for the currency used in the receipt',
        ),
        # XXX: Is this always mandatory?
    )
    related_receipts = models.ManyToManyField(
        'Receipt',
        verbose_name=_('related receipts'),
        blank=True,
    )

    objects = ReceiptManager()

    # TODO: Not implemented: optionals
    # TODO: methods to validate totals

    # These two values are stored in the receipt's batch. However, before the
    # receipt is assigned into a batch, this value should be used.
    receipt_type = models.ForeignKey(
        ReceiptType,
        related_name='receipts',
        verbose_name=_('receipt type'),
        on_delete=models.PROTECT,
    )
    point_of_sales = models.ForeignKey(
        PointOfSales,
        related_name='receipts',
        verbose_name=_('point of sales'),
        on_delete=models.PROTECT,
    )

    @property
    def formatted_number(self):
        return '{:04d}-{:08d}'.format(
            self.point_of_sales.number,
            self.receipt_number,
        )

    def validate(self, ticket=None):
        """
        Validates this receipt.

        This is a shortcut to :class:`~.ReceiptQuerySet`'s method of the same
        name. Calling this validates only this instance.
        """
        rv = Receipt.objects.filter(pk=self.pk).validate(ticket)
        # Since we're operating via a queryset, this instance isn't properly
        # updated:
        self.refresh_from_db()
        return rv

    def __repr__(self):
        return '<Receipt {}: {} {} for {}>'.format(
            self.pk,
            self.receipt_type,
            self.receipt_number,
            self.point_of_sales.owner,
        )

    def __str__(self):
        if self.receipt_number:
            return _('%(receipt_type)s %(receipt_number)s') % {
                'receipt_type': self.receipt_type,
                'receipt_number': self.formatted_number,
            }
        else:
            return _('Unnumbered %s') % self.receipt_type

    class Meta:
        ordering = ('issued_date',)
        verbose_name = _('receipt')
        verbose_name_plural = _('receipts')
        unique_together = (
            ('point_of_sales', 'receipt_type', 'receipt_number',)
        )
        # TODO: index_together...


class ReceiptPDFManager(models.Manager):

    def create_for_receipt(self, receipt, **kwargs):
        """
        Creates a ReceiptPDF object for a given receipt. Does not actually
        generate the related PDF file.

        :param Receipt receipt: The receipt for the PDF will be generated.
        """
        try:
            profile = TaxPayerProfile.objects.get(
                taxpayer__points_of_sales__receipts=receipt,
            )
        except TaxPayerProfile.DoesNotExist:
            raise Exception(
                'Cannot generate a PDF for taxpayer with no profile',
            )
        pdf = ReceiptPDF.objects.create(
            receipt=receipt,
            issuing_name=profile.issuing_name,
            issuing_address=profile.issuing_address,
            issuing_email=profile.issuing_email,
            vat_condition=profile.vat_condition,
            gross_income_condition=profile.gross_income_condition,
            sales_terms=profile.sales_terms,
            **kwargs
        )
        return pdf


class ReceiptPDF(models.Model):
    """
    Printable version of a receipt.

    Models all print-related data of a receipt and references generated PDF
    files.

    All `issuing`` fields contain data for the entity issuing the Receipt
    (these may change from one receipt to the next if, for example, the entity
    moved).
    """
    receipt = models.OneToOneField(
        Receipt,
        verbose_name=_('receipt'),
        on_delete=models.PROTECT,
    )
    pdf_file = models.FileField(
        verbose_name=_('pdf file'),
        upload_to='receipts',
        blank=True,
        null=True,
    )
    issuing_name = models.CharField(
        max_length=128,
        verbose_name=_('issuing name'),
    )
    issuing_address = models.TextField(
        _('issuing address'),
    )
    issuing_email = models.CharField(
        max_length=128,
        verbose_name=_('issuing email'),
        blank=True,
        null=True,
    )
    vat_condition = models.CharField(
        max_length=48,
        verbose_name=_('vat condition'),
        # IVA Responsable Inscripto
        # No responsable IVA
        # IVA Exento
        # A consumidor Final
    )
    gross_income_condition = models.CharField(
        max_length=48,
        verbose_name=_('gross income condition'),
    )
    client_name = models.CharField(
        max_length=128,
        verbose_name=_('client name'),
    )
    client_address = models.TextField(
        _('client address'),
    )
    sales_terms = models.CharField(
        max_length=48,
        verbose_name=_('sales terms'),
        help_text=_(
            'Should be something like "Cash", "Payable in 30 days", etc'
        ),
    )

    objects = ReceiptPDFManager()

    def _check_authorized(self):
        if not self.receipt.receipt_number:
            raise Exception(
                _('Cannot generate pdf for non-authorized receipt')
            )

    def save_pdf(self):
        """
        Save the receipt as a PDF related to this model.

        The related :class:`~.Receipt` should be validated first, of course.
        """
        self._check_authorized()

        with NamedTemporaryFile(suffix='.pdf') as file_:
            self.save_pdf_to(file_)
            self.pdf_file = File(file_, name='{}.pdf'.format(uuid.uuid4().hex))
            self.save()

    def save_pdf_to(self, file_):
        """
        Save the receipt as an actual PDF file into a custom location.

        The related :class:`~.Receipt` should be validated first, of course.
        """
        self._check_authorized()

        from . import pdf
        pdf.generate_receipt_pdf(self.receipt_id, file_)

    def __str__(self):
        return _('Receipt PDF for %s') % self.receipt_id

    class Meta:
        verbose_name = _('receipt pdf')
        verbose_name_plural = _('receipt pdfs')


class ReceiptEntry(models.Model):
    """
    An entry in a receipt.

    Each ReceiptEntry represents a line in printable version of a Receipt. You
    should generally have one instance per product or service.

    Note that each entry has a :class:`~.Vat` because a single Receipt can have
    multiple products with different :class:`~.VatType`.
    """

    receipt = models.ForeignKey(
        Receipt,
        related_name='entries',
        verbose_name=_('receipt'),
        on_delete=models.PROTECT,
    )
    description = models.CharField(
        max_length=128,
        verbose_name=_('description'),
    )
    quantity = models.PositiveSmallIntegerField(
        _('quantity'),
    )
    unit_price = models.DecimalField(
        _('unit price'),
        max_digits=15,
        decimal_places=2,
    )
    vat = models.ForeignKey(
        VatType,
        related_name='receipt_entries',
        verbose_name=_('vat'),
        null=True,
        on_delete=models.PROTECT,
    )

    @property
    def total_price(self):
        """The total price for this line (quantity * price)."""
        return self.quantity * self.unit_price

    class Meta:
        verbose_name = _('receipt entry')
        verbose_name_plural = _('receipt entries')


class Tax(models.Model):
    """A tax (type+amount) for a specific Receipt."""

    tax_type = models.ForeignKey(
        TaxType,
        verbose_name=_('tax type'),
        on_delete=models.PROTECT,
    )
    description = models.CharField(
        _('description'),
        max_length=80,
    )
    base_amount = models.DecimalField(
        _('base amount'),
        max_digits=15,
        decimal_places=2,
    )
    aliquot = models.DecimalField(
        _('aliquot'),
        max_digits=5,
        decimal_places=2,
    )
    amount = models.DecimalField(
        _('amount'),
        max_digits=15,
        decimal_places=2,
    )

    receipt = models.ForeignKey(
        Receipt,
        related_name='taxes',
        on_delete=models.PROTECT,
    )

    def compute_amount(self):
        """Auto-assign and return the total amount for this tax."""
        self.amount = self.base_amount * self.aliquot / 100
        return self.amount

    class Meta:
        verbose_name = _('tax')
        verbose_name_plural = _('taxes')


class Vat(models.Model):
    """A VAT (type+amount) for a specific Receipt."""

    vat_type = models.ForeignKey(
        VatType,
        verbose_name=_('vat type'),
        on_delete=models.PROTECT,
    )
    base_amount = models.DecimalField(
        _('base amount'),
        max_digits=15,
        decimal_places=2,
    )
    amount = models.DecimalField(
        _('amount'),
        max_digits=15,
        decimal_places=2,
    )

    receipt = models.ForeignKey(
        Receipt,
        related_name='vat',
        on_delete=models.PROTECT,
    )

    class Meta:
        verbose_name = _('vat')
        verbose_name_plural = _('vat')


class Validation(models.Model):
    """
    The validation result for a batch.

    The validation result for an attempt to validate a batch. Note that each
    Receipt inside the batch will have its own :class:`~.ReceiptValidation`.
    """
    RESULT_APPROVED = 'A'
    RESULT_REJECTED = 'R'
    RESULT_PARTIAL = 'P'

    processed_date = models.DateTimeField(
        _('processed date'),
    )
    result = models.CharField(
        _('result'),
        max_length=1,
        choices=(
            (RESULT_APPROVED, _('approved')),
            (RESULT_REJECTED, _('rejected')),
            (RESULT_PARTIAL, _('partial')),
        ),
    )

    # reprocessed?

    batch = models.ForeignKey(
        ReceiptBatch,
        related_name='validation',
        verbose_name=_('receipt batch'),
        on_delete=models.PROTECT,
    )

    class Meta:
        verbose_name = _('validation')
        verbose_name_plural = _('validations')


class Observation(models.Model):
    """
    An observation returned by AFIP.

    AFIP seems to assign re-used codes to Observation, so we actually store
    them as separate objects, and link to them from failed validations.
    """
    code = models.PositiveSmallIntegerField(
        _('code'),
    )
    message = models.CharField(
        _('message'),
        max_length=255,
    )

    class Meta:
        verbose_name = _('observation')
        verbose_name_plural = _('observations')


class ReceiptValidation(models.Model):
    """
    The validation for a single :class:`~.Receipt`.

    This contains all validation-related data for a receipt, including its CAE
    and the CAE expiration, unless validation has failed.

    The ``observation`` field may contain any data returned by AFIP regarding
    validation failure.
    """

    validation = models.ForeignKey(
        Validation,
        verbose_name=_('validation'),
        related_name='receipts',
        help_text=_(
            'The validation for the batch that produced this instance.'
        ),
        on_delete=models.PROTECT,
    )
    # TODO: replace this with a `successful` boolean field.
    result = models.CharField(
        _('result'),
        max_length=1,
        choices=(
            (Validation.RESULT_APPROVED, _('approved')),
            (Validation.RESULT_REJECTED, _('rejected')),
        ),
        help_text=_('Indicates whether the validation was succesful or not'),
    )
    cae = models.CharField(
        _('cae'),
        max_length=14,
        help_text=_('The CAE as returned by the AFIP'),
    )
    cae_expiration = models.DateField(
        _('cae expiration'),
        help_text=_('The CAE expiration as returned by the AFIP'),
    )
    observations = models.ManyToManyField(
        Observation,
        verbose_name=_('observations'),
        related_name='validations',
        help_text=_(
            'The observations as returned by the AFIP. These are generally '
            'present for failed validations.'
        ),
    )

    receipt = models.OneToOneField(
        Receipt,
        related_name='validation',
        verbose_name=_('receipt'),
        help_text=_('The Receipt for which this validation applies'),
        on_delete=models.PROTECT,
    )

    def __repr__(self):
        return '<{} {}: {} for Receipt {}>'.format(
            self.__class__.__name__,
            self.pk,
            self.result,
            self.receipt_id,
        )

    class Meta:
        verbose_name = _('receipt validation')
        verbose_name_plural = _('receipt validations')



from unittest.mock import call, MagicMock, patch

from django.test import TestCase

from django_afip import models
from testapp.testmain import fixtures


class ReceiptQuerySetTestCase(TestCase):

    def test_default_manager(self):
        self.assertIsInstance(
            models.Receipt.objects.all(),
            models.ReceiptQuerySet,
        )

    def test_validate(self):
        receipt = fixtures.ReceiptFactory()
        queryset = models.Receipt.objects.filter(pk=receipt.pk)
        ticket = MagicMock()

        with patch(
            'django_afip.models.ReceiptBatch.validate', spec=True,
        ) as mocked_validate:
            queryset.validate(ticket)

        self.assertEqual(mocked_validate.call_count, 1)
        self.assertEqual(mocked_validate.call_args, call(ticket))


class ReceiptManagerTestCase(TestCase):

    def test_default_manager(self):
        self.assertIsInstance(models.Receipt.objects, models.ReceiptManager)


class ReceiptTestCase(TestCase):

    def test_validate(self):
        receipt = fixtures.ReceiptFactory()
        ticket = MagicMock()

        with patch(
            'django_afip.models.ReceiptBatchManager.create', spec=True,
        ) as mocked_validate:
            receipt.validate(ticket)

        self.assertEqual(mocked_validate.call_count, 1)
        self.assertQuerysetEqual(
            mocked_validate.call_args[0][0],
            (receipt.pk,),
            lambda x: (x.pk),
        )
        self.assertEqual(mocked_validate.call_args[0][0].count(), 1)

        mocked_batch = mocked_validate()
        self.assertEqual(mocked_batch.validate.call_count, 1)
        self.assertEqual(mocked_batch.validate.call_args, call(ticket))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-06-29 18:35
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0017_receipt_issued_date'),
    ]

    operations = [
        migrations.AlterField(
            model_name='receipt',
            name='concept',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='receipts', to='afip.ConceptType', verbose_name='concept'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='currency',
            field=models.ForeignKey(help_text='Currency in which this receipt is issued.', on_delete=django.db.models.deletion.PROTECT, related_name='documents', to='afip.CurrencyType', verbose_name='currency'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='document_type',
            field=models.ForeignKey(help_text='The document type of the customer to whom this receipt is addressed', on_delete=django.db.models.deletion.PROTECT, related_name='receipts', to='afip.DocumentType', verbose_name='document type'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='point_of_sales',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='receipts', to='afip.PointOfSales', verbose_name='point of sales'),
        ),
        migrations.AlterField(
            model_name='receipt',
            name='receipt_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='receipts', to='afip.ReceiptType', verbose_name='receipt type'),
        ),
        migrations.AlterField(
            model_name='receiptbatch',
            name='point_of_sales',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='afip.PointOfSales', verbose_name='point of sales'),
        ),
        migrations.AlterField(
            model_name='receiptbatch',
            name='receipt_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='afip.ReceiptType', verbose_name='receipt type'),
        ),
        migrations.AlterField(
            model_name='receiptentry',
            name='receipt',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='entries', to='afip.Receipt', verbose_name='receipt'),
        ),
        migrations.AlterField(
            model_name='receiptentry',
            name='vat',
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, related_name='receipt_entries', to='afip.VatType', verbose_name='vat'),
        ),
        migrations.AlterField(
            model_name='receiptpdf',
            name='receipt',
            field=models.OneToOneField(on_delete=django.db.models.deletion.PROTECT, to='afip.Receipt', verbose_name='receipt'),
        ),
        migrations.AlterField(
            model_name='receiptvalidation',
            name='receipt',
            field=models.OneToOneField(help_text='The Receipt for which this validation applies', on_delete=django.db.models.deletion.PROTECT, related_name='validation', to='afip.Receipt', verbose_name='receipt'),
        ),
        migrations.AlterField(
            model_name='receiptvalidation',
            name='validation',
            field=models.ForeignKey(help_text='The validation for the batch that produced this instance.', on_delete=django.db.models.deletion.PROTECT, related_name='receipts', to='afip.Validation', verbose_name='validation'),
        ),
        migrations.AlterField(
            model_name='tax',
            name='receipt',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='taxes', to='afip.Receipt'),
        ),
        migrations.AlterField(
            model_name='tax',
            name='tax_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='afip.TaxType', verbose_name='tax type'),
        ),
        migrations.AlterField(
            model_name='validation',
            name='batch',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='validation', to='afip.ReceiptBatch', verbose_name='receipt batch'),
        ),
        migrations.AlterField(
            model_name='vat',
            name='receipt',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='vat', to='afip.Receipt'),
        ),
        migrations.AlterField(
            model_name='vat',
            name='vat_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='afip.VatType', verbose_name='vat type'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2016-02-03 02:56
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0011_receipt_entry_vat'),
    ]

    operations = [
        migrations.AlterField(
            model_name='taxpayerprofile',
            name='taxpayer',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='profile', to='afip.TaxPayer', verbose_name='taxpayer'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations
import django_afip.models


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AuthTicket',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('unique_id', models.IntegerField(verbose_name='id único', default=django_afip.models.AuthTicket.default_unique_id)),
                ('generated', models.DateTimeField(verbose_name='generado', default=django_afip.models.AuthTicket.default_generated)),
                ('expires', models.DateTimeField(verbose_name='vence', default=django_afip.models.AuthTicket.default_expires)),
                ('service', models.CharField(verbose_name='servicio', help_text='Servicio para el cual este ticket ha sido autorizado.', max_length=6)),
                ('token', models.TextField(verbose_name='token')),
                ('signature', models.TextField(verbose_name='firma')),
            ],
            options={
                'verbose_name': 'ticket de autorización',
                'verbose_name_plural': 'tickets de autorización',
            },
        ),
        migrations.CreateModel(
            name='ConceptType',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('code', models.CharField(verbose_name='código', max_length=3)),
                ('description', models.CharField(verbose_name='descripción', max_length=250)),
                ('valid_from', models.DateField(null=True, blank=True, verbose_name='válido desde')),
                ('valid_to', models.DateField(null=True, blank=True, verbose_name='válido hasta')),
            ],
            options={
                'verbose_name': 'tipo de concepto',
                'verbose_name_plural': 'tipos de concepto',
            },
        ),
        migrations.CreateModel(
            name='CurrencyType',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('code', models.CharField(verbose_name='código', max_length=3)),
                ('description', models.CharField(verbose_name='descripción', max_length=250)),
                ('valid_from', models.DateField(null=True, blank=True, verbose_name='válido desde')),
                ('valid_to', models.DateField(null=True, blank=True, verbose_name='válido hasta')),
            ],
            options={
                'verbose_name': 'tipo de moneda',
                'verbose_name_plural': 'tipos de moneda',
            },
        ),
        migrations.CreateModel(
            name='DocumentType',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('code', models.CharField(verbose_name='código', max_length=3)),
                ('description', models.CharField(verbose_name='descripción', max_length=250)),
                ('valid_from', models.DateField(null=True, blank=True, verbose_name='válido desde')),
                ('valid_to', models.DateField(null=True, blank=True, verbose_name='válido hasta')),
            ],
            options={
                'verbose_name': 'tipo de documento',
                'verbose_name_plural': 'tipos de documento',
            },
        ),
        migrations.CreateModel(
            name='Observation',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('code', models.PositiveSmallIntegerField(verbose_name='código')),
                ('message', models.CharField(verbose_name='mensaje', max_length=255)),
            ],
            options={
                'verbose_name': 'observación',
                'verbose_name_plural': 'observaciones',
            },
        ),
        migrations.CreateModel(
            name='PointOfSales',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('number', models.PositiveSmallIntegerField(verbose_name='número')),
                ('issuance_type', models.CharField(verbose_name='tipo de emisión', help_text='Indicates if thie POS emits using CAE and CAEA.', max_length=8)),
                ('blocked', models.BooleanField(verbose_name='bloqueado')),
                ('drop_date', models.DateField(null=True, blank=True, verbose_name='fecha de baja')),
            ],
            options={
                'verbose_name': 'punto de ventas',
                'verbose_name_plural': 'puntos de venta',
            },
        ),
        migrations.CreateModel(
            name='Receipt',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('document_number', models.BigIntegerField(verbose_name='número de documento', help_text='El número de documento del cliente para el cual este comprobante es emitido.')),
                ('receipt_number', models.PositiveIntegerField(null=True, blank=True, verbose_name='número de comprobante')),
                ('issued_date', models.DateField(verbose_name='fecha de emisión', help_text='Puede diverger hasta 5 días para bienes, o 10 en otros casos.')),
                ('total_amount', models.DecimalField(verbose_name='monto total', max_digits=15, help_text='Debe ser igual al monto no gravado + monto gravado + tributos + iva.', decimal_places=2)),
                ('net_untaxed', models.DecimalField(verbose_name='monto total no gravado', max_digits=15, help_text='El total para el cual impuestos no aplican. Para comprobantes tipo C, debe ser cero.', decimal_places=2)),
                ('net_taxed', models.DecimalField(verbose_name='monto total gravado', max_digits=15, help_text='EL total para el cual impuestos  aplican. Para comprobantes tipo C, esto es igual al subtotal.', decimal_places=2)),
                ('exempt_amount', models.DecimalField(verbose_name='monto exento', max_digits=15, help_text='Sólo para rubros que son exentos de impuestos. Para comprobantes tipo C, debe ser cero.', decimal_places=2)),
                ('service_start', models.DateField(null=True, blank=True, help_text='Fecha en la cual el servicio comenzó. No aplica a bienes.', verbose_name='fecha de comienzo de servicio')),
                ('service_end', models.DateField(null=True, blank=True, help_text='Fecha en la cual el servicio finalizó. No aplica a bienes.', verbose_name='fecha de fin de servicio')),
                ('expiration_date', models.DateField(null=True, blank=True, help_text='Fecha en la cual este comprobante expira. No aplica a bienes.', verbose_name='fecha de vencimiento del comprobante')),
                ('currency_quote', models.DecimalField(verbose_name='cotización de moneda', max_digits=10, help_text='Cotizacón del día para la moneda usada en el comprobante.', decimal_places=6)),
            ],
            options={
                'verbose_name': 'comprobante',
                'verbose_name_plural': 'comprobantes',
            },
        ),
        migrations.CreateModel(
            name='ReceiptBatch',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('point_of_sales', models.ForeignKey(verbose_name='punto de ventas', to='afip.PointOfSales', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name': 'lote de comprobantes',
                'verbose_name_plural': 'lotes de comprobantes',
            },
        ),
        migrations.CreateModel(
            name='ReceiptType',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('code', models.CharField(verbose_name='código', max_length=3)),
                ('description', models.CharField(verbose_name='descripción', max_length=250)),
                ('valid_from', models.DateField(null=True, blank=True, verbose_name='válido desde')),
                ('valid_to', models.DateField(null=True, blank=True, verbose_name='válido hasta')),
            ],
            options={
                'verbose_name': 'tipo de comprobante',
                'verbose_name_plural': 'tipos de comprobante',
            },
        ),
        migrations.CreateModel(
            name='ReceiptValidation',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('result', models.CharField(verbose_name='resultado', max_length=1, choices=[('A', 'aprobado'), ('R', 'rechazado'), ('P', 'parcial')])),
                ('cae', models.CharField(verbose_name='cae', max_length=14)),
                ('cae_expiration', models.DateField(verbose_name='expiración de cae')),
                ('observations', models.ManyToManyField(verbose_name='observaciones', to='afip.Observation', related_name='valiations')),
                ('receipt', models.OneToOneField(verbose_name='comprobante', to='afip.Receipt', related_name='validation', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name': 'validación de comprobante',
                'verbose_name_plural': 'validaciones de comprobante',
            },
        ),
        migrations.CreateModel(
            name='Tax',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('description', models.CharField(verbose_name='descripción', max_length=80)),
                ('base_amount', models.DecimalField(verbose_name='monto base', max_digits=15, decimal_places=2)),
                ('aliquot', models.DecimalField(verbose_name='alicuota', max_digits=5, decimal_places=2)),
                ('amount', models.DecimalField(verbose_name='monto', max_digits=15, decimal_places=2)),
                ('receipt', models.ForeignKey(related_name='taxes', to='afip.Receipt', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name': 'tributo',
                'verbose_name_plural': 'tributos',
            },
        ),
        migrations.CreateModel(
            name='TaxPayer',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('name', models.CharField(verbose_name='nombre', help_text='Un nombre amigable para reconocer a este contribuyente.', max_length=32)),
                ('key', models.FileField(null=True, verbose_name='clave', upload_to='')),
                ('certificate', models.FileField(null=True, verbose_name='certificado', upload_to='')),
                ('cuit', models.BigIntegerField(verbose_name='cuit')),
            ],
            options={
                'verbose_name': 'contribuyente',
                'verbose_name_plural': 'contribuyentes',
            },
        ),
        migrations.CreateModel(
            name='TaxType',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('code', models.CharField(verbose_name='código', max_length=3)),
                ('description', models.CharField(verbose_name='descripción', max_length=250)),
                ('valid_from', models.DateField(null=True, blank=True, verbose_name='válido desde')),
                ('valid_to', models.DateField(null=True, blank=True, verbose_name='válido hasta')),
            ],
            options={
                'verbose_name': 'tipo de tributo',
                'verbose_name_plural': 'tipos de tributo',
            },
        ),
        migrations.CreateModel(
            name='Validation',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('processed_date', models.DateTimeField(verbose_name='fecha procesado')),
                ('result', models.CharField(verbose_name='resultado', max_length=1, choices=[('A', 'aprobado'), ('R', 'rechazado'), ('P', 'parcial')])),
                ('batch', models.ForeignKey(verbose_name='lote de comprobantes', to='afip.ReceiptBatch', related_name='validation', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name': 'validación',
                'verbose_name_plural': 'validaciones',
            },
        ),
        migrations.CreateModel(
            name='Vat',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('base_amount', models.DecimalField(verbose_name='monto base', max_digits=15, decimal_places=2)),
                ('amount', models.DecimalField(verbose_name='monto', max_digits=15, decimal_places=2)),
                ('receipt', models.ForeignKey(related_name='vat', to='afip.Receipt', on_delete=models.CASCADE)),
            ],
            options={
                'verbose_name': 'iva',
                'verbose_name_plural': 'iva',
            },
        ),
        migrations.CreateModel(
            name='VatType',
            fields=[
                ('id', models.AutoField(serialize=False, auto_created=True, primary_key=True, verbose_name='ID')),
                ('code', models.CharField(verbose_name='código', max_length=3)),
                ('description', models.CharField(verbose_name='descripción', max_length=250)),
                ('valid_from', models.DateField(null=True, blank=True, verbose_name='válido desde')),
                ('valid_to', models.DateField(null=True, blank=True, verbose_name='válido hasta')),
            ],
            options={
                'verbose_name': 'tipo de iva',
                'verbose_name_plural': 'tipos de iva',
            },
        ),
        migrations.AddField(
            model_name='vat',
            name='vat_type',
            field=models.ForeignKey(verbose_name='tipo de iva', to='afip.VatType', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='tax',
            name='tax_type',
            field=models.ForeignKey(verbose_name='tipo de tributo', to='afip.TaxType', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='receiptvalidation',
            name='validation',
            field=models.ForeignKey(verbose_name='validación', to='afip.Validation', related_name='receipts', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='receiptbatch',
            name='receipt_type',
            field=models.ForeignKey(verbose_name='tipo de comprobante', to='afip.ReceiptType', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='receipt',
            name='batch',
            field=models.ForeignKey(null=True, blank=True, help_text='Los comprobantes son validados en lotes, así que debe ser asignado a unou antes de que sea posible validarlo.', verbose_name='lote de comprobantes', to='afip.ReceiptBatch', related_name='receipts', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='receipt',
            name='concept',
            field=models.ForeignKey(verbose_name='concepto', to='afip.ConceptType', related_name='receipts', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='receipt',
            name='currency',
            field=models.ForeignKey(verbose_name='moneda', to='afip.CurrencyType', related_name='documents', help_text='Moneda en la cual el comprobante es emitido.', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='receipt',
            name='document_type',
            field=models.ForeignKey(verbose_name='tipo de documento', to='afip.DocumentType', related_name='receipts', help_text='El tipo de documento del client para el cual este recibo es emitido.', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='receipt',
            name='point_of_sales',
            field=models.ForeignKey(to='afip.PointOfSales', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='receipt',
            name='receipt_type',
            field=models.ForeignKey(to='afip.ReceiptType', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='receipt',
            name='related_receipts',
            field=models.ManyToManyField(to='afip.Receipt', blank=True, db_constraint='comprobantes relacionados'),
        ),
        migrations.AddField(
            model_name='pointofsales',
            name='owner',
            field=models.ForeignKey(verbose_name='dueño', to='afip.TaxPayer', on_delete=models.CASCADE),
        ),
        migrations.AddField(
            model_name='authticket',
            name='owner',
            field=models.ForeignKey(verbose_name='dueño', to='afip.TaxPayer', related_name='auth_tickets', on_delete=models.CASCADE),
        ),
        migrations.AlterUniqueTogether(
            name='receipt',
            unique_together=set([('point_of_sales', 'receipt_type', 'receipt_number')]),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='tax',
            name='amount',
            field=models.DecimalField(decimal_places=2, max_digits=15, verbose_name='cantidad'),
        ),
        migrations.AlterField(
            model_name='vat',
            name='amount',
            field=models.DecimalField(decimal_places=2, max_digits=15, verbose_name='cantidad'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.7 on 2017-05-29 20:12
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0015_receiptentry__amount_to_quantity'),
    ]

    operations = [
        migrations.AlterField(
            model_name='receipt',
            name='issued_date',
            field=models.DateField(help_text='Can diverge up to 5 days for goods, or 10 days otherwise', verbose_name='issued date'),
        ),
        migrations.AlterField(
            model_name='receiptentry',
            name='quantity',
            field=models.PositiveSmallIntegerField(verbose_name='quantity'),
        ),
        migrations.AlterField(
            model_name='receiptpdf',
            name='sales_terms',
            field=models.CharField(help_text='Should be something like "Cash", "Payable in 30 days", etc', max_length=48, verbose_name='sales terms'),
        ),
        migrations.AlterField(
            model_name='receiptvalidation',
            name='cae',
            field=models.CharField(help_text='The CAE as returned by the AFIP', max_length=14, verbose_name='cae'),
        ),
        migrations.AlterField(
            model_name='receiptvalidation',
            name='cae_expiration',
            field=models.DateField(help_text='The CAE expiration as returned by the AFIP', verbose_name='cae expiration'),
        ),
        migrations.AlterField(
            model_name='receiptvalidation',
            name='observations',
            field=models.ManyToManyField(help_text='The observations as returned by the AFIP. These are generally present for failed validations.', related_name='validations', to='afip.Observation', verbose_name='observations'),
        ),
        migrations.AlterField(
            model_name='receiptvalidation',
            name='receipt',
            field=models.OneToOneField(help_text='The Receipt for which this validation applies', on_delete=django.db.models.deletion.CASCADE, related_name='validation', to='afip.Receipt', verbose_name='receipt'),
        ),
        migrations.AlterField(
            model_name='receiptvalidation',
            name='result',
            field=models.CharField(choices=[('A', 'approved'), ('R', 'rejected')], help_text='Indicates whether the validation was succesful or not', max_length=1, verbose_name='result'),
        ),
        migrations.AlterField(
            model_name='receiptvalidation',
            name='validation',
            field=models.ForeignKey(help_text='The validation for the batch that produced this instance.', on_delete=django.db.models.deletion.CASCADE, related_name='receipts', to='afip.Validation', verbose_name='validation'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2015-12-12 18:20
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0007_auto_20151212_1754'),
    ]

    operations = [
        migrations.AlterField(
            model_name='receiptpdf',
            name='receipt',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, to='afip.Receipt', verbose_name='receipt'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0003_vat_tax_digits'),
    ]

    operations = [
        migrations.AlterField(
            model_name='tax',
            name='amount',
            field=models.DecimalField(verbose_name='cantidad', max_digits=15, decimal_places=2),
        ),
        migrations.AlterField(
            model_name='vat',
            name='amount',
            field=models.DecimalField(verbose_name='cantidad', max_digits=15, decimal_places=2),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9.2 on 2016-06-04 17:17
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0013_taxpayer_is_sandboxed'),
    ]

    operations = [
        migrations.AlterField(
            model_name='receiptvalidation',
            name='result',
            field=models.CharField(choices=[('A', 'approved'), ('R', 'rejected')], max_length=1, verbose_name='result'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.9 on 2015-12-14 18:36
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0008_auto_20151212_1820'),
    ]

    operations = [
        migrations.AlterField(
            model_name='receiptpdf',
            name='issuing_email',
            field=models.CharField(blank=True, max_length=128, null=True, verbose_name='issuing email'),
        ),
        migrations.AlterField(
            model_name='taxpayerprofile',
            name='issuing_email',
            field=models.CharField(blank=True, max_length=128, null=True, verbose_name='issuing email'),
        ),
        migrations.AlterField(
            model_name='taxpayerprofile',
            name='issuing_name',
            field=models.CharField(max_length=128, verbose_name='issuing name'),
        ),
    ]



from datetime import date, datetime

from factory import SubFactory
from factory.django import DjangoModelFactory

from django_afip import models


class ConceptTypeFactory(DjangoModelFactory):
    class Meta:
        model = models.ConceptType

    code = '1'
    description = 'Producto'
    valid_from = date(2010, 9, 17)


class DocumentTypeFactory(DjangoModelFactory):
    class Meta:
        model = models.DocumentType

    code = '96'
    description = 'DNI'
    valid_from = date(2008, 7, 25)


class CurrencyTypeFactory(DjangoModelFactory):
    class Meta:
        model = models.CurrencyType

    code = 'PES'
    description = 'Pesos Argentinos'
    valid_from = date(2009, 4, 3)


class ReceiptTypeFactory(DjangoModelFactory):
    class Meta:
        model = models.ReceiptType

    code = '11'
    description = 'Factura C'
    valid_from = date(2011, 3, 30)


class TaxPayerFactory(DjangoModelFactory):
    class Meta:
        model = models.TaxPayer

    name = 'John Smith'
    cuit = '20329642330'
    is_sandboxed = True


class PointOfSalesFactory(DjangoModelFactory):
    class Meta:
        model = models.PointOfSales

    number = 1
    issuance_type = 'CAE'
    blocked = False
    owner = SubFactory(TaxPayerFactory)


class ReceiptFactory(DjangoModelFactory):
    class Meta:
        model = models.Receipt

    concept = SubFactory(ConceptTypeFactory)
    document_type = SubFactory(DocumentTypeFactory)
    document_number = 33445566
    issued_date = datetime(2017, 5, 15)
    total_amount = 100
    net_untaxed = 0
    net_taxed = 100
    exempt_amount = 0
    currency = SubFactory(CurrencyTypeFactory)
    currency_quote = 1
    receipt_type = SubFactory(ReceiptTypeFactory)
    point_of_sales = SubFactory(PointOfSalesFactory)



from datetime import date
from tempfile import NamedTemporaryFile

from django.contrib.auth.models import User
from django.core.files.base import File

from django_afip import models


def taxpayer(key=None):
    taxpayer = models.TaxPayer.objects.create(
        name='Test Taxpayer',
        cuit=20329642330,
        is_sandboxed=True,
    )

    if key:
        with NamedTemporaryFile(suffix='.key') as file_:
            file_.write(key)
            taxpayer.key = File(file_, name='test.key')
            taxpayer.save()

    return taxpayer


def superuser():
    return User.objects.create_superuser('test', 'test@example.com', '123')


def receipt(
        document_type=96,
        with_tax=True,
        with_vat=True,
        receipt_type=6,
     ):
    """
    Return a dummy mocked-receipt.

    This function is here for convenience only and has no special magic other
    than creating a receipt with valid Vat and Tax attributes.
    """
    total_amount = 100
    if with_tax:
        total_amount += 9
    if with_vat:
        total_amount += 21
    receipt = models.Receipt.objects.create(
        concept=models.ConceptType.objects.get(code=1),
        document_type=models.DocumentType.objects.get(
            code=document_type,
        ),
        document_number='203012345',
        issued_date=date.today(),
        total_amount=total_amount,
        net_untaxed=0,
        net_taxed=100,
        exempt_amount=0,
        currency=models.CurrencyType.objects.get(code='PES'),
        currency_quote=1,

        receipt_type=models.ReceiptType.objects.get(code=receipt_type),
        point_of_sales=models.PointOfSales.objects.first(),
    )
    if with_vat:
        models.Vat.objects.create(
            vat_type=models.VatType.objects.get(code=5),
            base_amount=100,
            amount=21,
            receipt=receipt,
        )
    if with_tax:
        models.Tax.objects.create(
            tax_type=models.TaxType.objects.get(code=3),
            base_amount=100,
            aliquot=9,
            amount=9,
            receipt=receipt,
        )
    return receipt



from django.apps import AppConfig

from . import monkeys


class AfipConfig(AppConfig):
    name = 'django_afip'
    label = 'afip'
    verbose_name = 'AFIP'

    def ready(self):
        monkeys.patch_https_for_afip()
